import pytest

from mlserver import types, __version__


def test_live(rest_client):
    endpoint = "/v2/health/live"
    response = rest_client.get(endpoint)

    assert response.status_code == 200


def test_ready(rest_client):
    endpoint = "/v2/health/ready"
    response = rest_client.get(endpoint)

    assert response.status_code == 200


def test_model_ready(rest_client, sum_model):
    endpoint = f"/v2/models/{sum_model.name}/versions/{sum_model.version}/ready"
    response = rest_client.get(endpoint)

    assert response.status_code == 200


def test_metadata(rest_client):
    endpoint = "/v2"
    response = rest_client.get(endpoint)

    metadata = types.MetadataServerResponse.parse_obj(response.json())

    assert metadata.name == "mlserver"
    assert metadata.version == __version__
    assert metadata.extensions == []


def test_model_metadata(rest_client, sum_model_settings):
    endpoint = f"v2/models/{sum_model_settings.name}"
    response = rest_client.get(endpoint)

    metadata = types.MetadataModelResponse.parse_obj(response.json())

    assert metadata.name == sum_model_settings.name
    assert metadata.platform == sum_model_settings.platform
    assert metadata.versions == sum_model_settings.versions
    assert metadata.inputs == sum_model_settings.inputs


@pytest.mark.parametrize(
    "model_name,model_version", [("sum-model", "v1.2.3"), ("sum-model", None)]
)
def test_infer(rest_client, inference_request, model_name, model_version):
    endpoint = f"/v2/models/{model_name}/infer"
    if model_version is not None:
        endpoint = f"/v2/models/{model_name}/versions/{model_version}/infer"
    response = rest_client.post(endpoint, json=inference_request.dict())

    assert response.status_code == 200

    prediction = types.InferenceResponse.parse_obj(response.json())
    assert len(prediction.outputs) == 1
    assert prediction.outputs[0].data == [21]


def test_infer_error(rest_client, inference_request):
    endpoint = "/v2/models/my-model/versions/v0/infer"
    response = rest_client.post(endpoint, json=inference_request.dict())

    assert response.status_code == 400
    assert response.json()["error"] == "Model my-model with version v0 not found"


def test_model_repository_index(rest_client, repository_index_request):
    endpoint = "/v2/repository/index"
    response = rest_client.post(endpoint, json=repository_index_request.dict())

    assert response.status_code == 200

    models = response.json()
    assert len(models) == 1


def test_model_repository_unload(rest_client, sum_model_settings):
    endpoint = f"/v2/repository/models/{sum_model_settings.name}/unload"
    response = rest_client.post(endpoint)

    assert response.status_code == 200

    model_metadata = rest_client.get(f"/v2/models/{sum_model_settings.name}")
    assert model_metadata.status_code == 400


def test_model_repository_load(rest_client, sum_model_settings):
    rest_client.post(f"/v2/repository/models/{sum_model_settings.name}/unload")

    endpoint = f"/v2/repository/models/{sum_model_settings.name}/load"
    response = rest_client.post(endpoint)

    assert response.status_code == 200

    model_metadata = rest_client.get(f"/v2/models/{sum_model_settings.name}")
    assert model_metadata.status_code == 200


def test_model_repository_load_error(rest_client, sum_model_settings):
    endpoint = "/v2/repository/models/my-model/load"
    response = rest_client.post(endpoint)

    assert response.status_code == 400
    assert response.json()["error"] == "Model my-model not found"



import pytest
import uuid

from mlserver.settings import ModelSettings, ModelParameters
from mlserver.types import MetadataTensor

from ..fixtures import SumModel


@pytest.mark.parametrize("ready", [True, False])
async def test_ready(data_plane, model_registry, ready):
    model_settings = ModelSettings(
        name="sum-model-2", parameters=ModelParameters(version="v1.2.3")
    )
    new_model = SumModel(model_settings)
    await model_registry.load(new_model)

    new_model.ready = ready

    all_ready = await data_plane.ready()

    assert all_ready == ready


@pytest.mark.parametrize("ready", [True, False])
async def test_model_ready(data_plane, sum_model, ready):
    sum_model.ready = ready
    model_ready = await data_plane.model_ready(sum_model.name, sum_model.version)

    assert model_ready == ready


@pytest.mark.parametrize(
    "server_name,server_version,extensions",
    [(None, None, None), ("my-server", "v2", ["foo", "bar"])],
)
async def test_metadata(settings, data_plane, server_name, server_version, extensions):
    if server_name is not None:
        settings.server_name = server_name

    if server_version is not None:
        settings.server_version = server_version

    if extensions is not None:
        settings.extensions = extensions

    metadata = await data_plane.metadata()

    assert metadata.name == settings.server_name
    assert metadata.version == settings.server_version
    assert metadata.extensions == settings.extensions


@pytest.mark.parametrize(
    "platform,versions,inputs",
    [
        (None, None, None),
        ("sklearn", ["sklearn/0.22.3"], None),
        (
            "xgboost",
            ["xgboost/1.1.0"],
            [MetadataTensor(name="input-0", datatype="FP32", shape=[6, 7])],
        ),
    ],
)
async def test_model_metadata(
    sum_model_settings, data_plane, platform, versions, inputs
):
    if platform is not None:
        sum_model_settings.platform = platform

    if versions is not None:
        sum_model_settings.versions = versions

    if inputs is not None:
        sum_model_settings.inputs = inputs

    metadata = await data_plane.model_metadata(
        name=sum_model_settings.name, version=sum_model_settings.parameters.version
    )

    assert metadata.name == sum_model_settings.name
    assert metadata.platform == sum_model_settings.platform
    assert metadata.versions == sum_model_settings.versions
    assert metadata.inputs == sum_model_settings.inputs


async def test_infer(data_plane, sum_model, inference_request):
    prediction = await data_plane.infer(
        payload=inference_request, name=sum_model.name, version=sum_model.version
    )

    assert len(prediction.outputs) == 1
    assert prediction.outputs[0].data == [21]


async def test_infer_generates_uuid(data_plane, sum_model, inference_request):
    inference_request.id = None
    prediction = await data_plane.infer(
        payload=inference_request, name=sum_model.name, version=sum_model.version
    )

    assert prediction.id is not None
    assert prediction.id == str(uuid.UUID(prediction.id))



from ..settings import ModelSettings
from ..registry import MultiModelRegistry
from ..repository import ModelRepository
from ..errors import ModelNotFound
from ..types import (
    RepositoryIndexRequest,
    RepositoryIndexResponse,
    RepositoryIndexResponseItem,
    State,
)


class ModelRepositoryHandlers:
    def __init__(self, repository: ModelRepository, model_registry: MultiModelRegistry):
        self._repository = repository
        self._model_registry = model_registry

    async def index(self, payload: RepositoryIndexRequest) -> RepositoryIndexResponse:
        all_model_settings = await self._repository.list()

        repository_items = []
        for model_settings in all_model_settings:
            index_item = await self._to_item(model_settings)
            if payload.ready:
                # TODO: If filtering by ready, we could ready directly from the
                # active model registry
                if index_item.state == State.READY:
                    repository_items.append(index_item)
            else:
                repository_items.append(index_item)

        return RepositoryIndexResponse(__root__=repository_items)

    async def _to_item(
        self, model_settings: ModelSettings
    ) -> RepositoryIndexResponseItem:
        item = RepositoryIndexResponseItem(
            name=model_settings.name,
            state=State.UNKNOWN,
            reason="",
        )

        item.state = await self._get_state(model_settings)

        if model_settings.parameters:
            item.version = model_settings.parameters.version

        return item

    async def _get_state(self, model_settings: ModelSettings) -> State:
        try:
            model = await self._model_registry.get_model(model_settings.name)
            if model.ready:
                return State.READY
        except ModelNotFound:
            return State.UNAVAILABLE

        return State.UNKNOWN

    async def load(self, name: str) -> bool:
        model_settings = await self._repository.find(name)

        # TODO: Move to separate method
        model_class = model_settings.implementation
        model = model_class(model_settings)  # type: ignore

        await self._model_registry.load(model)

        return True

    async def unload(self, name: str) -> bool:
        await self._model_registry.unload(name)

        return True



from fastapi.responses import Response

from ..types import (
    MetadataModelResponse,
    MetadataServerResponse,
    InferenceRequest,
    InferenceResponse,
    RepositoryIndexRequest,
    RepositoryIndexResponse,
)
from ..handlers import DataPlane, ModelRepositoryHandlers

from .utils import to_status_code


class Endpoints:
    """
    Implementation of REST endpoints.
    These take care of the REST/HTTP-specific things and then delegate the
    business logic to the internal handlers.
    """

    def __init__(self, data_plane: DataPlane):
        self._data_plane = data_plane

    async def live(self) -> Response:
        is_live = await self._data_plane.live()
        return Response(status_code=to_status_code(is_live))

    async def ready(self) -> Response:
        is_ready = await self._data_plane.ready()
        return Response(status_code=to_status_code(is_ready))

    async def model_ready(self, model_name: str, model_version: str = None) -> Response:
        is_ready = await self._data_plane.model_ready(model_name, model_version)
        return Response(status_code=to_status_code(is_ready))

    async def metadata(self) -> MetadataServerResponse:
        return await self._data_plane.metadata()

    async def model_metadata(
        self, model_name: str, model_version: str = None
    ) -> MetadataModelResponse:
        return await self._data_plane.model_metadata(model_name, model_version)

    async def infer(
        self,
        payload: InferenceRequest,
        model_name: str,
        model_version: str = None,
    ) -> InferenceResponse:
        return await self._data_plane.infer(payload, model_name, model_version)


class ModelRepositoryEndpoints:
    def __init__(self, handlers: ModelRepositoryHandlers):
        self._handlers = handlers

    async def index(self, payload: RepositoryIndexRequest) -> RepositoryIndexResponse:
        return await self._handlers.index(payload)

    async def load(self, model_name: str) -> Response:
        loaded = await self._handlers.load(name=model_name)
        return Response(status_code=to_status_code(loaded))

    async def unload(self, model_name: str) -> Response:
        unloaded = await self._handlers.unload(name=model_name)
        return Response(status_code=to_status_code(unloaded))



import orjson

from typing import Any
from fastapi import Request


class ORJSONRequest(Request):
    """
    Custom request class which uses `orjson`.
    """

    async def json(self) -> Any:
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = orjson.loads(body)
        return self._json



# generated by datamodel-codegen:
#   filename:  dataplane.yaml

from typing import List, Optional, Union

from pydantic import BaseModel, Field


class MetadataServerResponse(BaseModel):
    name: str
    version: str
    extensions: List[str]


class MetadataServerErrorResponse(BaseModel):
    error: str


class MetadataTensor(BaseModel):
    name: str
    datatype: str
    shape: List[int]


class MetadataModelErrorResponse(BaseModel):
    error: str


class Parameters(BaseModel):
    pass


class TensorData(BaseModel):
    __root__: List[Union["TensorData", Union[float, str, bool]]] = Field(
        ..., title="tensor_data"
    )

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, idx):
        return self.__root__[idx]

    def __len__(self):
        return len(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, dict):
            return v.get("__root__", [])
        return v


class RequestOutput(BaseModel):
    name: str
    parameters: Optional["Parameters"] = None


class ResponseOutput(BaseModel):
    name: str
    shape: List[int]
    datatype: str
    parameters: Optional["Parameters"] = None
    data: "TensorData"


class InferenceResponse(BaseModel):
    model_name: str
    model_version: Optional[str] = None
    id: Optional[str] = None
    parameters: Optional["Parameters"] = None
    outputs: List["ResponseOutput"]


class InferenceErrorResponse(BaseModel):
    error: Optional[str] = None


class MetadataModelResponse(BaseModel):
    name: str
    versions: Optional[List[str]] = None
    platform: str
    inputs: Optional[List["MetadataTensor"]] = None
    outputs: Optional[List["MetadataTensor"]] = None


class RequestInput(BaseModel):
    name: str
    shape: List[int]
    datatype: str
    parameters: Optional["Parameters"] = None
    data: "TensorData"


class InferenceRequest(BaseModel):
    id: Optional[str] = None
    parameters: Optional["Parameters"] = None
    inputs: List["RequestInput"]
    outputs: Optional[List["RequestOutput"]] = None


TensorData.update_forward_refs()



from .version import __version__
from .server import MLServer
from .model import MLModel
from .settings import Settings, ModelSettings

__all__ = ["__version__", "MLServer", "MLModel", "Settings", "ModelSettings"]



import os

from typing import List

from .types import RequestInput
from .settings import ModelSettings
from .errors import InvalidModelURI

try:
    import numpy as np
except ImportError:
    # TODO: Log warning message
    pass

NP_DTYPES = {
    "BOOL": "bool",
    "UINT8": "uint8",
    "UINT16": "uint16",
    "UINT32": "uint32",
    "UINT64": "uint64",
    "INT8": "int8",
    "INT16": "int16",
    "INT32": "int32",
    "INT64": "int64",
    "FP16": "float16",
    "FP32": "float32",
    "FP64": "float64",
    "BYTES": "byte",
}


async def get_model_uri(
    settings: ModelSettings, wellknown_filenames: List[str] = []
) -> str:
    if not settings.parameters:
        raise InvalidModelURI(settings.name)

    model_uri = settings.parameters.uri

    if not model_uri:
        raise InvalidModelURI(settings.name)

    if os.path.isfile(model_uri):
        return model_uri

    if os.path.isdir(model_uri):
        # If model_uri is a folder, search for a well-known model filename
        for fname in wellknown_filenames:
            model_path = os.path.join(model_uri, fname)
            if os.path.isfile(model_path):
                return model_path

        # If none, return the folder
        return model_uri

    # Otherwise, the uri is neither a file nor a folder
    raise InvalidModelURI(settings.name, model_uri)


def to_ndarray(request_input: RequestInput) -> "np.ndarray":
    dtype = to_dtype(request_input.datatype)
    data = getattr(request_input.data, "__root__", request_input.data)

    model_data = np.array(data, dtype=dtype)

    # TODO: Check if reshape not valid
    return model_data.reshape(request_input.shape)


def to_dtype(datatype: str) -> "np.dtype":
    # TODO: Validate datatype earlier (in Pydantic)
    dtype = NP_DTYPES[datatype]
    return np.dtype(dtype)



# generated by datamodel-codegen:
#   filename:  model_repository.yaml

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field


class RepositoryIndexRequest(BaseModel):
    ready: Optional[bool] = None


class State(Enum):
    UNKNOWN = "UNKNOWN"
    READY = "READY"
    UNAVAILABLE = "UNAVAILABLE"
    LOADING = "LOADING"
    UNLOADING = "UNLOADING"


class RepositoryIndexResponseItem(BaseModel):
    name: str
    version: Optional[str] = None
    state: "State"
    reason: str


class RepositoryIndexResponse(BaseModel):
    __root__: List["RepositoryIndexResponseItem"] = Field(
        ..., title="repository_index_response"
    )

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, idx):
        return self.__root__[idx]

    def __len__(self):
        return len(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, dict):
            return v.get("__root__", [])
        return v


class RepositoryIndexErrorResponse(BaseModel):
    error: Optional[str] = None


class RepositoryLoadErrorResponse(BaseModel):
    error: Optional[str] = None


class RepositoryUnloadErrorResponse(BaseModel):
    error: Optional[str] = None



from typing import List, Optional
from pydantic import BaseSettings, PyObject

from .version import __version__
from .types import MetadataTensor

ENV_PREFIX_SETTINGS = "MLSERVER_"
ENV_PREFIX_MODEL_SETTINGS = "MLSERVER_MODEL_"


class Settings(BaseSettings):
    class Config:
        env_prefix = ENV_PREFIX_SETTINGS

    debug: bool = True

    # Model repository folder
    model_repository_root: str = "."

    # Server metadata
    server_name: str = "mlserver"
    server_version: str = __version__
    extensions: List[str] = []

    # Server settings
    host: str = "0.0.0.0"
    http_port: int = 8080
    grpc_port: int = 8081
    grpc_workers: int = 10


class ModelParameters(BaseSettings):
    """
    Parameters that apply only to a particular instance of a model.
    This can include things like model weights.
    The main difference with respect to ModelSettings is that parameters can
    change on each instance (e.g. each version) of the model.
    """

    class Config:
        env_prefix = ENV_PREFIX_MODEL_SETTINGS

    uri: Optional[str] = None
    version: Optional[str] = None
    format: Optional[str] = None


class ModelSettings(BaseSettings):
    class Config:
        env_prefix = ENV_PREFIX_MODEL_SETTINGS

    name: str

    # Model metadata
    platform: str = ""
    versions: Optional[List[str]] = []
    inputs: Optional[List[MetadataTensor]] = []
    outputs: Optional[List[MetadataTensor]] = []

    # Custom model class implementation
    implementation: PyObject = "mlserver.model.MLModel"  # type: ignore

    # Model parameters are meant to be set directly by the MLServer runtime.
    # However, it's also possible to override them manually.
    parameters: Optional[ModelParameters] = None



import pytest

from typing import Optional

from mlserver.errors import ModelNotFound
from mlserver.registry import MultiModelRegistry
from mlserver.handlers import ModelRepositoryHandlers
from mlserver.settings import ModelSettings
from mlserver.types import RepositoryIndexRequest, State


async def test_index(
    model_repository_handlers: ModelRepositoryHandlers,
    repository_index_request: RepositoryIndexRequest,
    sum_model_settings: ModelSettings,
):
    repo_index = list(await model_repository_handlers.index(repository_index_request))

    assert len(repo_index) == 1
    assert repo_index[0].name == sum_model_settings.name
    assert (
        repo_index[0].version == sum_model_settings.parameters.version  # type: ignore
    )
    assert repo_index[0].state == State.READY


async def test_index_unavailable_model(
    model_repository_handlers: ModelRepositoryHandlers,
    repository_index_request: RepositoryIndexRequest,
    sum_model_settings: ModelSettings,
):
    await model_repository_handlers.unload(sum_model_settings.name)
    repo_index = list(await model_repository_handlers.index(repository_index_request))

    assert len(repo_index) == 1
    assert repo_index[0].name == sum_model_settings.name
    assert (
        repo_index[0].version == sum_model_settings.parameters.version  # type: ignore
    )
    assert repo_index[0].state == State.UNAVAILABLE


@pytest.mark.parametrize("ready,expected", [(None, 1), (True, 0), (False, 1)])
async def test_index_filter_ready(
    model_repository_handlers: ModelRepositoryHandlers,
    repository_index_request: RepositoryIndexRequest,
    sum_model_settings: ModelSettings,
    ready: Optional[bool],
    expected: int,
):
    await model_repository_handlers.unload(sum_model_settings.name)

    repository_index_request.ready = ready
    repo_index = list(await model_repository_handlers.index(repository_index_request))

    assert len(repo_index) == expected


async def test_unload(
    model_repository_handlers: ModelRepositoryHandlers,
    model_registry: MultiModelRegistry,
    sum_model_settings: ModelSettings,
):
    await model_repository_handlers.unload(sum_model_settings.name)

    with pytest.raises(ModelNotFound):
        await model_registry.get_model(sum_model_settings.name)


async def test_unload_not_found(
    model_repository_handlers: ModelRepositoryHandlers,
):
    with pytest.raises(ModelNotFound):
        await model_repository_handlers.unload("not-existing")


async def test_load_not_found(
    model_repository_handlers: ModelRepositoryHandlers,
):
    with pytest.raises(ModelNotFound):
        await model_repository_handlers.load("not-existing")
